/****************************************************************************/
/* file       :  GU_1.c                                                     */
/*                                                                          */
/* include    :  ALBERT                                                     */
/*                                                                          */
/* description:  Laplce Solver                                              */
/*                                                                          */
/* date       :  Sep 16  2004                                               */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*    author  : Pyo, Jae-Hong                                               */
/*                                                                          */
/*              Deaprtment of Mathematics                                   */
/*              Changwon National University                                */
/*                                                                          */
/*    email   : jhpyo@cjangwon.ac.kr                                        */
/*              http://www.math.purdue.edu/~pjh                             */
/*                                                                          */
/****************************************************************************/
#include <albert.h>
static const QUAD            *quad;
static const FE_SPACE        *fe_space;
static const int             n_refine;
static DOF_MATRIX            *MATRIX;
static DOF_REAL_VEC          *Force;
static DOF_REAL_VEC          *Sol;

struct op_info
{
  REAL_D         Lambda[DIM+1];
  REAL           det;
};

static void init_element(const EL_INFO *el_info, const QUAD *quad[3], void *ud)
{
  struct op_info *info = ud;

  info->det = el_grd_lambda(el_info, info->Lambda);

  return;
}

void init_dof_admin(MESH *mesh)
{
  FUNCNAME("init_dof_admin");

  int               degree = 1;
  const BAS_FCTS    *lagrange;

  lagrange   = get_lagrange(degree);
  TEST_EXIT(lagrange)("no lagrange BAS_FCTS\n");
  fe_space = get_fe_space(mesh, lagrange->name,   nil, lagrange);

  return;
}

static void mesh_refine_fct(const EL_INFO *el_info)
{
  el_info->el->mark = n_refine;
}


const REAL (*make_loc_gg(const EL_INFO *el_info, const QUAD *quad, 
		  int iq, void *ud))[DIM+1]
{
  struct op_info *info = ud;
  int            i, j, k;
  static REAL    LALt[DIM+1][DIM+1];

  for (i = 0; i <= DIM; i++)
    for (j = i; j <= DIM; j++)
    {
      for (LALt[i][j] = k = 0; k < DIM_OF_WORLD; k++)
	LALt[i][j] += info->Lambda[i][k]*info->Lambda[j][k];
      LALt[i][j] *=info->det;
      LALt[j][i] = LALt[i][j];
    }

  return((const REAL (*)[DIM+1]) LALt);
}

REAL make_loc_ss(const EL_INFO *el_info,const QUAD *quad,int iq, void *ud)
{
  struct op_info *info = ud;

  return(info->det);
}

static void assemble_GG_mat_fct(const EL_INFO *el_info)
{
  FUNCNAME("assemble_GG_mat_fct");
  static const EL_MATRIX_INFO *matrix_info = nil;
  static const REAL           **loc_mat;
  static int                  dim_loc_mat;
  static void                 *loc_info;
  static int                  *um_dof;
  static S_CHAR               *bound;

  if (!matrix_info)
  {
    OPERATOR_INFO  o_info = {nil};
    dim_loc_mat = MATRIX->fe_space->bas_fcts->n_bas_fcts;

    o_info.row_fe_space   = o_info.col_fe_space = MATRIX->fe_space;
    o_info.quad[2]        = quad;
    o_info.init_element   = init_element;
    o_info.LALt           = make_loc_gg;
    o_info.LALt_pw_const  = false;
    o_info.LALt_symmetric = true;
    o_info.use_get_bound  = true;
    o_info.user_data      = MEM_ALLOC(1, struct op_info);
    o_info.fill_flag      = CALL_LEAF_EL|FILL_COORDS;

    matrix_info           = fill_matrix_info(&o_info, nil);
    loc_info              = matrix_info->fill_info;

    if (um_dof)
      {
	free(um_dof); free(bound);
      }
    um_dof                = MEM_ALLOC(dim_loc_mat, int);
    bound                 = MEM_ALLOC(dim_loc_mat, S_CHAR);
  }
  MATRIX->fe_space->bas_fcts->get_dof_indices(el_info->el,
			      MATRIX->fe_space->admin,um_dof);
  loc_mat  = matrix_info->el_matrix_fct(el_info, loc_info);
  MATRIX->fe_space->bas_fcts->get_bound(el_info,bound);
  add_element_matrix(MATRIX,1.0, dim_loc_mat, dim_loc_mat,
		     um_dof,um_dof,loc_mat,bound);
  /*
    add_element_matrix(MATRIX,CON_ALPHA, dim_loc_mat, dim_loc_mat,
    um_dof,um_dof,loc_mat,nil); Nuemann Boundary */
}

static REAL ext_u_fct(const REAL_D x)
{
  return(x[0]*(1-x[0])*x[1]*(1-x[1]));
}

static REAL force_fct(const REAL_D x)
{
  return(2.0*x[0]*(1-x[0])+2.0*x[1]*(1-x[1]));
}

int main(int argc, char **argv)
{
  FUNCNAME("main");
  GLTOOLS_WINDOW    win_exact, win_numerical;
  MESH              *mesh;
  char              filename[100];
 
  init_parameters(0, "hw_2.dat");
  GET_PARAMETER(1, "macro file name", "%s", filename);
  GET_PARAMETER(1, "global refinements", "%d", &n_refine);

  mesh = GET_MESH("hw_1", init_dof_admin, nil);
  mesh->preserve_coarse_dofs = 0;
  read_macro(mesh, filename, nil);

  mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS|FILL_BOUND, mesh_refine_fct);
  refine(mesh);

  dof_compress(mesh);

  MATRIX   = get_dof_matrix("MATRIX", fe_space);
  Force    = get_dof_real_vec("Force", fe_space);
  Sol      = get_dof_real_vec("Sol", fe_space);

  quad   = get_quadrature(DIM, 2*fe_space->bas_fcts->degree);

  clear_dof_matrix(MATRIX);
  mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS|FILL_BOUND, assemble_GG_mat_fct);
  dof_set(0.0, Force);            dof_set(0.0, Sol);

  L2scp_fct_bas(force_fct, quad, Force);
  dirichlet_bound(nil, Force, Sol, nil);

  oem_solve_s(MATRIX, Force, nil, Sol, 2, 1.e-10, 1, 10, 30000, 5);
 
  dof_set(0.0, Force);             interpol(ext_u_fct, Force);
  win_exact = open_gltools_window("Exact U", "400x400+0", nil, mesh, 1);
  win_numerical = open_gltools_window("Numerical U", "400x400+400", nil, mesh, 1);
  gltools_drv(win_exact, Force, 1, 0);
  gltools_drv(win_numerical, Sol, 1, 0);
 
  return(0);
}
